/************************************************************************/
/*  mssieve -	Provide a sieve to catch and discard MiniSEED data	*/
/*		that is shorter than the specified timespan.		*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Berkeley Seismological Laboratory				*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 2002-2014 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static const char sccsid[] = "$Id: mssieve.c,v 1.2 2014/10/07 00:06:24 doug Exp $ ";
#endif

#include <stdio.h>
#include <string.h>
#include "qlib2.h"
#include "version.h"

#include "externals.h"

char *syntax[] = {
"%s version " VERSION,
"%s - Filter and discard MiniSEED data segments shorter than specified timespan or lower clock quality.",
"%s [-s timespan] [-q percent] [-o outfile | -n] [-a] [-r] [-G tol]",
"    [-D j|m] [-d n] [-h] [input_file_1]",
"    where:",
"	-s time_span",
"		    Minimumm timespan of data to keep.",
"		    Discard any segment of data shorter than this time.",
"	-q percent  Minimum clock quality to keep.  Default quality is 0.",
"	-o outfile  Optional name of output file.",
"		    If not specified or specified as '-', write to stdout.", 
"	-n	    No data output.  Equivalent to -o /dev/null.",
"	-a	    Append output to file instead of overwriting file.",
"	-r	    Reverse operations.  Output the shorter segments,",
"		    and discard the longer segments.",
"	-G TOL	    Explicit tolerance for time gaps in ticks (1/10 msec).",
"		    Default is " STRING(TTOL_MULT) " of data sample interval.",
"	-D j | m    Display dates in julian or month and day format.",
"	-d n	    Debug option N.",
"		    1 = debug record header and record cache",
"	-h	    Help - prints syntax message.",
"	input_file",
"		    Data file with original data.  If no file is specified,",
"		    read from stdin.",
NULL };

#include <stdlib.h>
#include <strings.h>
#include <math.h>

int append_cache (char *p, int n);
int fwrite_cache (FILE *fp);
int flush_cache (FILE *fp);
int tol (int rate, int rate_mult);
int parse_cmdline (int *pargc, char ***pargv);

/************************************************************************/
/*  main:   main program						*/
/************************************************************************/
int main (int argc, char **argv)
{
    FILE *fpin, *fpout;
    DATA_HDR *hdr = NULL;
    char *mseed = NULL;
    INT_TIME istart, exptime, to_time;
    int blksize;
    int blkno = 0;
    int cacheing_data = 0;
    double bdif, ispan;

    /*	Initialization.							*/
    output_file= NULL;		/* empty initial output directory.	*/
    fpin = stdin;		/* assume input to come from stdin.	*/
    info = stdout;
    date_fmt = JULIAN_FMT;

    parse_cmdline (&argc, &argv);

    if (input_file) {
	if ((fpin = fopen (input_file,"r")) == NULL) {
	    fprintf (info, "Unable to open input file %s\n", input_file);
	    exit(1);
	}
    }
    else {
	input_file = "<stdin>";
	fpin = stdin;
    }

    if ((! no_output) && (output_file && strcmp(output_file, "-") != 0)) {
	if ((fpout = fopen (output_file,outmode)) == NULL) {
	    fprintf (info, "Unable to open output file %s\n", output_file);
	    exit(1);
	}
    }
    else {
	info = stderr;
	output_file = "<stdout>";
	fpout = stdout;
    }

    while (( hdr=NULL, mseed=NULL, blksize = read_ms_record (&hdr, &mseed, fpin)) > 0) {
	int seconds, usecs;
	BS *bs;

	/* If sieving on clock quality, discard if:	*/
	/* a.  have blockette 1001 and			*/
	/* b.  clock quality not high enough.		*/
	if (min_clock_quality > 0 && (bs=find_blockette(hdr, 1001))) {
	    BLOCKETTE_1001 *b = (BLOCKETTE_1001 *) bs->pb;
	    /* These are all byte values, so I can ignore wordorder.	*/
	    int clock_quality = b->clock_quality;
	    /* Discard block if clock quality is lower than specified min_clock_quality. */
	    if (clock_quality < min_clock_quality) continue;	
	}
	
	/* Initialize interval start and expected time for first record in file.    */
	if (blkno++ == 0) {
	    istart = hdr->begtime;
	    exptime = hdr->begtime;
	}

	/* Compute timespan of record. */
	time_interval2 (hdr->num_samples, hdr->sample_rate, hdr->sample_rate_mult, &seconds, &usecs);
	to_time = add_time (hdr->begtime,seconds,usecs);

	/* Compute the span from interval start through current record. */
	ispan = tdiff(to_time,istart);

	/* Compute gap between current and previous record. */
	bdif = tdiff(hdr->begtime, exptime);
	if (debug(DEBUG_CACHE)) {
	    fprintf (info, "Record %s.%s.%s.%s %s to ", 
		     hdr->station_id, hdr->network_id, hdr->channel_id, hdr->location_id, 
		     time_to_str(hdr->begtime, date_fmt));
	    fprintf (info, "%s gap=%.6lf ispan=%.6lf ", time_to_str(to_time, date_fmt),
		     bdif/USECS_PER_SEC, ispan/USECS_PER_SEC);
	}

	/* Check to see if records are contiguous. */
	if (fabs(bdif) > DIHUGE || fabs(bdif) >= tol(hdr->sample_rate, hdr->sample_rate_mult)) {
	    /* Data not continuous.  */
	    if (! reversed) {
		if (debug(DEBUG_CACHE)) fprintf (info, "flush_cache, ");
		if (! flush_cache(fpout)) exit(1);
	    }
	    else {
		cacheing_data = 0;
		if (debug(DEBUG_CACHE)) fprintf (info, "fwrite_cache, ");
		if (! fwrite_cache(fpout)) exit(1);
	    }
	    istart = hdr->begtime;
	    exptime = hdr->begtime;
	    ispan = tdiff(to_time,istart);
	}

	/* Determine whether we exceed the minimum segment or whether to cache the record. */
	if (ispan >= keep_usecs) {
	    /* We exceed the timespan criteria. */
	    if (! reversed) {
		if (cacheing_data) {
		    /* Write all data in cache */
		    if (debug(DEBUG_CACHE)) fprintf (info, "fwrite_cache, ");
		    if ((! no_output) && (! fwrite_cache (fpout))) exit(1);
		    cacheing_data = 0;
		}
		/* Output this record.		*/
		if (debug(DEBUG_CACHE)) fprintf (info, "fwrite_record\n");
		if ((! no_output) && fwrite(mseed,hdr->blksize,1,fpout) != 1) {
		    fprintf (info, "Error: writing output file\n");
		    exit(1);
		}
	    }
	    else {
		if (cacheing_data) {
		    /* Flush data in cache */
		    if (debug(DEBUG_CACHE)) fprintf (info, "flush_cache, ");
		    if (! flush_cache(fpout)) exit(1);
		    cacheing_data = 0;
		}
		if (debug(DEBUG_CACHE)) fprintf (info, "flush_record\n");
	    }
	}
	else {
	    /* Cache the record for possible use. */
	    if (debug(DEBUG_CACHE)) fprintf (info, "append_cache\n");
	    if (! append_cache (mseed, hdr->blksize)) exit(1);
	    cacheing_data = 1;
	}
	if (hdr) free_data_hdr(hdr);
	if (mseed) free(mseed);
	exptime = to_time;
    }

    /* Cleanup before exit. */
    if (hdr) free_data_hdr(hdr);
    if (mseed) free (mseed);
    fclose (fpin);
    if (fpout) fclose (fpout);
    return(0);
}

/************************************************************************/
/*  tol:								*/
/*	Return tolerance for continuous data for this stream.		*/
/*	Default value (if not overridded by cmdline argument)		*/
/*	is based on sample rate.					*/
/************************************************************************/
int tol
   (int		rate,		/* sample rate in qlib convention.	*/
    int		rate_mult)	/* sample rate_mult in qlib convention.	*/
{
    if (stream_tol >= 0) 
	return (stream_tol);
    else
	return (TOL(rate,rate_mult));
}

/************************************************************************/
/*  Cache routines for MiniSEED records.				*/
/*  Cache is just one large malloc/realloc buffer.			*/
/************************************************************************/

#define	CACHEINCR   (4096*1024)
static int cachesize = 0;
static int cacheused = 0;
static char *cache = NULL;

/************************************************************************/
/*  append_cache:	Append buffer to cache.				*/
/*			Return 1 on success, 0 on failure.		*/
/************************************************************************/
int append_cache (char *p, int n)
{
    while (cacheused+n > cachesize) {
	cache = (char *)realloc(cache,cachesize+CACHEINCR);
	if (cache == NULL) {
	    fprintf (info, "Error: mallocing mseed cache\n");
	    return (0);
	}
	cachesize += CACHEINCR;
    }
    memcpy (cache+cacheused, p, n);
    cacheused += n;
    return (1);
}

/************************************************************************/
/*  fwrite_cache:   Fwrite contents of cache.				*/
/*		    Return 1 on success, 0 on failure.			*/
/************************************************************************/
int fwrite_cache (FILE *fp)
{
    int n = 1;	
    if (fp && cacheused) n = fwrite (cache, cacheused, 1, fp);
    cacheused = 0;
    return (n);
}

/************************************************************************/
/*  flush_cache:    Flush cache.					*/
/*		    Return 1 on success, 0 on failure.			*/
/************************************************************************/
int flush_cache (FILE *fp)
{
    cacheused = 0;
    return (1);
}

